﻿/*
Copyright 2013 Antanas Veiverys www.veiverys.com

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.IO;
using System.IO.Ports;
using System.Linq;
using System.Reflection;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

namespace ComfilePiDrivers
{
    public class EnhancedSerialPort : SerialPort
    {
        [DllImport("MonoPosixHelper", SetLastError = true)]
        static extern bool poll_serial(int fd, out int error, int timeout);

        [DllImport("libc")]
        static extern IntPtr strerror(int errnum);

        static void ThrowIOException()
        {
            int errnum = Marshal.GetLastWin32Error();
            string error_message = Marshal.PtrToStringAnsi(strerror(errnum));

            throw new IOException(error_message);
        }

        public EnhancedSerialPort() : base()
        {
        }

        public EnhancedSerialPort(IContainer container) : base(container)
        {
        }

        public EnhancedSerialPort(string portName) : base(portName)
        {
        }

        public EnhancedSerialPort(string portName, int baudRate)
            : base(portName, baudRate)
        {
        }

        public EnhancedSerialPort(string portName, int baudRate, Parity parity)
            : base(portName, baudRate, parity)
        {
        }

        public EnhancedSerialPort(string portName, int baudRate, Parity parity, int dataBits)
            : base(portName, baudRate, parity, dataBits)
        {
        }

        public EnhancedSerialPort(string portName, int baudRate, Parity parity, int dataBits, StopBits stopBits)
            : base(portName, baudRate, parity, dataBits, stopBits)
        {
        }

        // private member access via reflection
        int fd;
        FieldInfo disposedFieldInfo;
        object data_received;

        Thread _thread;
        volatile bool _stopThread;

        public void Start()
        {
            if (Environment.OSVersion.Platform == PlatformID.Unix)
            {
                FieldInfo fieldInfo = BaseStream.GetType().GetField("fd", BindingFlags.Instance | BindingFlags.NonPublic);
                fd = (int)fieldInfo.GetValue(BaseStream);
                disposedFieldInfo = BaseStream.GetType().GetField("disposed", BindingFlags.Instance | BindingFlags.NonPublic);
                fieldInfo = typeof(SerialPort).GetField("data_received", BindingFlags.Instance | BindingFlags.NonPublic);
                data_received = fieldInfo.GetValue(this);

                _thread = new Thread(EventThreadFunction);
                _thread.IsBackground = true;
                _stopThread = false;
                _thread.Start();
            }
        }

        bool Poll(Stream stream, int timeout)
        {
            CheckDisposed(stream);
            if (IsOpen == false)
            {
                throw new Exception("port is closed");
            }
            int error;

            bool poll_result = poll_serial(fd, out error, ReadTimeout);
            if (error == -1)
            {
                ThrowIOException();
            }
            return poll_result;
        }

        public void Stop()
        {
            if (Environment.OSVersion.Platform == PlatformID.Unix)
            {
                if (_thread != null)
                {
                    _stopThread = true;
                    _thread.Join();
                    _thread = null;
                }
            }
        }

        private void EventThreadFunction()
        {
            while(!_stopThread)
            {
                if (IsOpen)
                {
                    try
                    {
                        var _stream = BaseStream;
                        if (_stream == null)
                        {
                            return;
                        }
                        if (Poll(_stream, ReadTimeout))
                        {
                            OnDataReceived(null);
                        }
                    }
                    catch
                    {
                        return;
                    }
                }

                Thread.Sleep(1);
            } 
        }

        void OnDataReceived(SerialDataReceivedEventArgs args)
        {
            ((SerialDataReceivedEventHandler)Events[data_received])?.Invoke(this, args);
        }

        void CheckDisposed(Stream stream)
        {
            bool disposed = (bool)disposedFieldInfo.GetValue(stream);
            if (disposed)
            {
                throw new ObjectDisposedException(stream.GetType().FullName);
            }
        }
    }
}
